#!/usr/bin/perl 
################################################################################
## YAFP - Yet another filter proxy                                             #
##                                                                             #
## This is based upon a proxy written by Randal L. Schwartz which is licensed  #
## under "the same terms as Perl itself" - As Perl itself is licensed under    #
## GPL _or_ Artistic License I've choosen the GPL for this software.           #
## So here comes the disclaimer:                                               #
##                                                                             #
## Copyright (C) 2001,2002 Andreas Gohr <a.gohr@web.de>                        #
##                                                                             #
## This program is free software; you can redistribute it and/or               #
## modify it under the terms of the GNU General Public License as              #
## published by the Free Software Foundation; either version 2 of              #
## the License, or (at your option) any later version.                         #
##                                                                             #
## See COPYING for details                                                     #
################################################################################
$|=1;
$VERSION = "0.1.1";

&setup_defaults();
&setup_signals();


### Look for configfiles
my $cfound = 0;
$cfound = 1 if (eval {require "/etc/yafp.conf";});
$cfound = 1 if (eval {require "/etc/yafp/yafp.conf";});
$cfound = 1 if (eval {require $ENV{'HOME'}."/.yafp.conf";});
$cfound = 1 if (eval {require "yafp.conf";});
die ("e No configfile found") unless $cfound;

&readconfigs();

### create PID file
open (PID,">$PIDFILE") or die("e Could not open PID file for writing");
print PID $$;
print PID "\n";
close PID;

### main

&main();
exit 0;

################################################################################ 
## subs

sub main {                      # return void
  use HTTP::Daemon;
  my %kids;
  &debug(2,"i running YAFP version ".$VERSION);
  &checkversions();
  my $master = HTTP::Daemon->new(LocalPort => $PORT, LocalAddr => $HOST)
       or die "Cannot create master: $!";
  &debug(2,"i master is ".$master->url);

  ## fork the right number of children
  for (1..$SLAVE_COUNT) {
    $kids{&fork_a_slave($master)} = "slave";
  }
  while(1){                             #Loop forever:
    my $pid = wait;
    my $was = delete ($kids{$pid}) || "?unknown?";
    
    if ($was eq "slave") {      # oops, lost a slave
      &debug(64,"* child $pid terminated");
      sleep 1;                  # don't replace it right away (avoid thrash)
      $kids{&fork_a_slave($master)} = "slave";
    }
  }
}
 
sub setup_signals {             # return void
  $SIG{__WARN__} = sub {
    &tosyslog('warning',join(" ",@_)) if($SYSLOG);
    warn &prefix(@_);
  };
  $SIG{__DIE__} = sub {
    &tosyslog('err',join(" ",@_)) if($SYSLOG);
    die &prefix(@_);
  };
  $SIG{INT} = $SIG{TERM} = sub {
    my $sig = shift;
    $SIG{$sig} = 'IGNORE';
    kill $sig, 0;               # kill the whole processgroup
    unlink($PIDFILE);
    &debug (32, "! catched $sig - terminating");
    exit 0;
  };
  $SIG{HUP} = sub {
    my $sig = shift;
    &debug (32, "! catched $sig - sending USR1 to all childs");
    kill USR1, 0;
  };
  $SIG{USR1} = sub {
    my $sig = shift;
    &debug (32, "! catched $sig - reloading configs");
    &readconfigs();             #reread the configs
  };
}
 
sub fork_a_slave {              # return int (pid)
  my $master = shift;           # HTTP::Daemon
 
  my $pid;
  defined ($pid = fork) or die "e Cannot fork: $!";
  &child_does($master) unless $pid;
  return $pid;
}
 
sub child_does {                # return void
  my $master = shift;           # HTTP::Daemon
  &debug(64,"* child started");

  while(1){
    if($LOCKSOCKET){
      flock($master, 2);          # LOCK_EX
      &debug(64, "* child has lock");
    }
    my $slave = $master->accept or die "accept: $!";
    if($LOCKSOCKET){
      &debug(64, "* child releasing lock");
      flock($master, 8);          # LOCK_UN
    }
    $slave->autoflush(1);
    &debug(8, ": connect from ". $slave->peerhost.':'.$slave->peerport);
    &handle_connection($slave); # closes $slave at right time
  }
  # this shouldn't be reached  ever
  &debug(64,"* child terminating");
  exit 0;
}


sub handle_connection{          # return void
  use HTTP::Request; 
  my $handle = shift;           # HTTP::Daemon::ClientConn (Socket)

  my $old_fh = select(SOCK);
  $|=1;
  select($old_fh);

  #Get request from browser
  my $request = $handle->get_request;                           # HTTP::Request
  defined($request) or die "e bad request ".$handle->reason;

  my $url = $request->uri();    # URI
  
  #only support http scheme:
  if ($url->scheme ne 'http'){
    my $res = HTTP::Response->new(403, "Forbidden");
    &debug(1,"e bad URL scheme: ".$url->scheme);
    $res->content(&errorhtml("only HTTP is supported",
                             "YAFP supports only http:// scheme URLs! You should set up
                              your Browser in a way that YAFP is only used for HTTP."));
    $handle->send_response($res);
    close $handle;
    return; # void
  }
  debug(4,"< $url");
  # check filterconditions based on URL:
  my $filter  = 1;
  my $cookies = 1;
  $_ = $url;
  if (&$is_accepturl){
    $filter  = 0;
    $cookies = 0;
  }elsif(&$is_cookieurl){
    $cookies = 0;
  }
  debug(128," filter: $filter cookies: $cookies");
  # filter Headers FIXME put together with above statement
  if ($filter){
    $request->remove_header(qw(User-Agent From Referer));
  }
  if ($cookies){
    $request->remove_header(qw(User-Agent From Referer Cookie));
    #FIXME send Wafer
  }

  #set up useragent:
  use LWP::UserAgent;
  BEGIN { *LWP::UserAgent::redirect_ok = sub { 0 } } #Never follow redirects
  my $agent = LWP::UserAgent->new;
  $agent->timeout($TIMEOUT);
  if ($USER_AGENT ne ''){
    $agent->agent($USER_AGENT);
  }else{
    $agent->agent("yafp/$VERSION " . $agent->agent);
  }
  $agent->env_proxy;

  #now we need to fetch it either direct to browser or to $content;
  my $content='';
  my $headerssent;
  my $received;
  my $expected;
  my $res = $agent->request($request,sub{
                              my ($chunk,$res)= @_ ;

                              $received += length($chunk);
                              unless (defined $expected) {
                                $expected = $res->content_length || 0;
                              }
                              debug(256,"% $received of $expected bytes");
                              if (($res->content_type =~ m/text\/html/) and
                                  (($res->content_encoding eq '') or
                                   ($res->content_encoding=~ m/gzip/) )){
                                $content .= $chunk;
                              }else{
                                unless(defined($headerssent)){
                                  $handle->send_status_line(); #HTTP/1.1 200 OK
                                  print $handle $res->headers->as_string();
                                  $handle->send_crlf();
                                  $headerssent = 1;
                                }
                                print $handle $chunk;
                              }
                            });
  debug(512,$res->headers->as_string());

  if ($res->code != 200){
    $handle->send_response($res);
    close $handle;
    return; #void
  }
  debug(16,"? ".$res->content_type);
  if ($headerssent){
    close $handle;
    return; #void
  }

  if ($res->content_encoding=~ m/gzip/){
    use Compress::Zlib;
    $content = Compress::Zlib::memGunzip($content);
  }
  #still here? so theres content to filter
  if ($filter){
    #Kill Cookies
    $res->remove_header(qw(Set-Cookie)) if $cookies;
    $content = &filterads($content);
  }
  if ($res->content_encoding=~ m/gzip/){
    use Compress::Zlib;
    $content = Compress::Zlib::memGzip($content);
  }
  $res->content($content);
  $res->content_length(length $content);
  $handle->send_response($res);
  close $handle;
  return; #void

}

sub filterads($){     #return string $content
  use HTML::TreeBuilder;
  my $content = $_[0]; #html content
  my $tree = HTML::TreeBuilder->new();
  $tree->ignore_unknown(0);
  $tree->warn(1);
  $tree->ignore_ignorable_whitespace(0);
  $tree->parse($content);
  $tree->eof();

  #all JavascriptEventhandlers:
  my @JSevents =  qw (onAbort onBlur onChange onClick onDblClick onError
                      onFocus onKeydown onKeypress onKeyup onMousedown
                      onMousemove onMouseout onMouseover onMouseUp onReset
                      onSelect onSubmit onLoad onUnLoad);

  if ($FILTER_SCRIPT){
    #Kill all SCRIPTs and events
    while($element = $tree->look_down("_tag","script")){
      $element->replace_with();
    }
    foreach my $event (@JSevents){
      while($element = $tree->look_down(sub {defined($_[0]->attr("$event"))})){
        $element->attr("$event",undef);
      }
    }
  }elsif ($FILTER_ONLOAD){
    #kill only load events 
    foreach my $event ("onLoad","onUnLoad"){
      while($element = $tree->look_down(sub {defined($_[0]->attr("$event"))})){
        $element->attr("$event",undef);
      }
    }
  }

  if ($FILTER_SIZE){
    foreach my $size (@BANNERSIZES){
      my ($w,$h) = split("x",$size);
      while($element = $tree->look_down("_tag","img",
                                        "width",$w,
                                        "height",$h)){
        $element->replace_with($AD_STRING);
      }
    }
  }

  if ((!$FILTER_SCRIPT) and ($FILTER_POPUP or $FILTER_STATUS)){
    #kill all unwanted script functions 
    foreach my $element ($tree->look_down("_tag","script")){
      my $script = $element->as_HTML;
      $script =~ s/(\w+\.)*open/delete/g if ($FILTER_POPUP);
      $script =~ s/(\w+\.)*popup/delete/g if ($FILTER_POPUP);
      $script =~ s/(\w+\.)*status/delete/g if ($FILTER_STATUS);
      $script =~ s/(\w+\.)*defaultStatus/unusedvariable/g  if ($FILTER_STATUS);
      $element->replace_with(HTML::Element->new('~literal','text' => $script));
    }
    #filter from events
    foreach my $event (@JSevents){
      foreach my $element ($tree->look_down(sub {defined($_[0]->attr("$event"))})){
        my $script = $element->attr("$event");
        $script =~ s/(\w+\.)*open/delete/g if ($FILTER_POPUP);
        $script =~ s/(\w+\.)*popup/delete/g if ($FILTER_POPUP);
        $script =~ s/(\w+\.)*status/delete/g if ($FILTER_STATUS);
        $script =~ s/(\w+\.)*defaultStatus/unusedvariable/g  if ($FILTER_STATUS);
        $element->attr("$event",$script);
      }
    }  
  }

  #Kill all tags that request a disallowed URL:
  while($element = $tree->look_down(sub{$_[0]->tag() =~ m/(img|iframe|layer|embed)/i},
                                    sub{$_ = $_[0]->attr('src'); return &$is_blockurl})){
    $element->replace_with($AD_STRING);
  }
  while($element = $tree->look_down("_tag","applet",
                                    sub{$_ = $_[0]->attr('code'); return &$is_blockurl})){
    $element->replace_with($AD_STRING);
  }
  while($element = $tree->look_down("_tag","applet",
                                    sub{$_ = $_[0]->attr('codebase'); return &$is_blockurl})){
    $element->replace_with($AD_STRING);
  }
  while($element = $tree->look_down("_tag","object",
                                    sub{$_ = $_[0]->attr('data'); return &$is_blockurl})){
    $element->replace_with($AD_STRING);
  }
  while($element = $tree->look_down("_tag","object",
                                    sub{$_ = $_[0]->attr('codebase'); return &$is_blockurl})){
    $element->replace_with($AD_STRING);
  }

  #done with parsing via HTML::BuildTree  
  $content = $tree->as_HTML();
  $tree->delete();

  #Apply patterns
  study $content;
  while (my ($patt,$repl)=each(%patterns)) {
    if ($content =~ s/$patt/$repl/igs){
      &debug(128,"~| ".$patt." >> ".$repl);
    }
  }
  
  return $content;  
}

sub readconfigs(){ # Return void;
# read config files and build regexp-function-references
  my @temp;
  my $line;
  my $patt;
  my $repl;
  my @accepturls;
  my @blockurls;
  my @cookieurls;
  undef(%patterns);
  
  # Never filter these URLs!
  if (open (ACCEPTURLS,$FILE_NOFILTER)){
    @temp = <ACCEPTURLS>;
    close(ACCEPTURLS);
    foreach $line (@temp){
      $line =~ s/#.*$//;
      $line =~ s/^\s+//;
      $line =~ s/\s+$//;
      next if ($line eq '');
      $line =~ s/[.\@\%\${}\/]/\\$&/g;
      $line =~ s/\*/.*?/g;
      push (@accepturls, $line);
    }
    #now build a matching sub (see the perl cookbook chapter 6.10)
    my $expr = join('||', map { "m/\$accepturls[$_]/io" } 0..$#accepturls);
    $is_accepturl = eval "sub {s/^http:\\/\\///; return $expr; }";
    die("An Error occured on building the accepturl patterns") if $@;
    &debug(2,"i loaded ".scalar(@accepturls)." accepturls");
  }else{
    $is_accepturl = eval "sub { return 0; }"; #This matches nothing
    &debug(1,"e Could not open '$FILE_NOFILTER'");
  }
  
  # Don't remove cookies from these URLs:
  if (open (COOKIEURLS,$FILE_COOKIE)){
    @temp = <COOKIEURLS>;
    close(COOKIEURLS);
    foreach $line (@temp){
      $line =~ s/#.*$//;
      $line =~ s/^\s+//;
      $line =~ s/\s+$//;
      next if ($line eq '');
      $line =~ s/[.\@\%\${}\/]/\\$&/g;
      $line =~ s/\*/.*?/g;
      push (@cookieurls, $line);
    }
    #now build a matching sub (see the perl cookbook chapter 6.10)
    my $expr = join('||', map { "m/\$cookieurls[$_]/io" } 0..$#cookieurls);
    $is_cookieurl = eval "sub { s/^http:\\/\\///; return $expr; }";
    die("An Error occured on building the cookieurl patterns") if $@;
    &debug(2,"i loaded ".scalar(@cookieurls)." cookieurls");
  }else{
    $is_accepturl = eval "sub { return 0; }"; #This matches nothing
    &debug(1,"e Could not open '$FILE_COOKIE'");
  }
  
  # Block these URLs!
  if (open (BLOCKURLS,$FILE_BLOCKURLS)){
    @temp = <BLOCKURLS>;
    close(BLOCKURLS);
    foreach $line (@temp){
      $line =~ s/#.*$//;
      $line =~ s/^\s+//;
      $line =~ s/\s+$//;
      next if ($line eq '');
      $line =~ s/[.\@\%\${}\/]/\\$&/g;
      $line =~ s/\*/.*?/g;
      push (@blockurls, $line);
    }
    #now build a matching sub (see the perl cookbook chapter 6.10)
    my $expr = join('||', map { "m/\$blockurls[$_]/io" } 0..$#blockurls);
    $is_blockurl = eval "sub { s/^http:\\/\\///; return $expr; }"; #This will be used in filterads()
    die("An Error occured on building the blockurl patterns") if $@;
    &debug(2,"i loaded ".scalar(@blockurls)." blockurls");    
  }else{
    $is_blockurl = eval "sub { return 0; }"; #This matches nothing
    &debug(1,"e Could not open '$FILE_BLOCKURLS'");
  }
  
  # Which Patterns should be replaced?
  if (open (PATTERNS,$FILE_PATTERNS)){
    @temp = <PATTERNS>;
    close(PATTERNS);
    foreach $line (@temp){
      $line =~ s/#.*$//;
      $line =~ s/^\s+//;
      $line =~ s/\s+$//;
      next if ($line eq '');
      ($patt,$repl)=split("\t",$line,2);
      $patt =~ s/[.\@\%\${}\/]/\\$&/g;
      $patt =~ s/\*/.*?/g;
      $patterns{$patt}=$repl;
    }
    &debug(2,"i loaded ".scalar(keys %patterns)." patterns");
  }else{
    &debug(1,"e Could not open '$FILE_PATTERNS'");
  }
}


sub errorhtml{
  my $title = shift;
  my $msg   = shift;
  return "
  <HTML>
  <HEAD>
    <TITLE>$title</TITLE>
  </HEAD>
  <BODY>
    <H3>$title</H3>
    <P>$msg</P>
    <HR SIZE=\"1\">
    <P ALIGN=\"right\">Yet Another Filter Proxy $VERSION</P>
  </BODY>
  </HTML>";
end
}


sub checkversions{
  my $version = $HTTP::Daemon::VERSION;
  debug(2,"i HTTP::Daemon is version  $version"); 
  $version = $LWP::UserAgent::VERSION;
  debug(2,"i LWP::UserAgent is version  $version"); 
  $version = $Compress::Zlib::VERSION;
  debug(2,"i Compress::Zlib is version  $version");
  $version = $Sys::Syslog::VERSION;
  debug(2,"i Sys::Syslog is version  $version");
  $version = $HTML::TreeBuilder::VERSION;
  debug(2,"i HTML::TreeBuilder is version $version");
  debug(1,"e Your HTML::TreeBuilder Module is out of date!".
          "\nThis may render some websites completely unusable!") if ($version < 3.11);
}

sub prefix {
  my @return;
  my @now = localtime();
  $now[5] = $now[5]+1900;
  my @mname = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
  foreach my $line (@_){
    push(@return,sprintf("[%s %2d %02d:%02d:%02d] [%d] %s",
         $mname[$now[4]],$now[3],$now[2],$now[1],$now[0],${$},$line));
  }
  return @return;
}

sub debug {           # return void
  my $lvl = shift; #Debuglevel of the message
  my $msg = shift; #debug message
  if (($DEBUGLEVEL & $lvl)){
    print STDERR prefix("$msg\n");
    if ($SYSLOG){
      my $syslvl;
      $syslvl = 'err' if (1 & $lvl);
      $syslvl = 'warning' if (2 & $lvl);
      $syslvl = 'notice' if (12 & $lvl);
      $syslvl = 'info' if (48 & $lvl);
      $syslvl = 'debug' unless (defined($syslvl));
      tosyslog($syslvl,$msg);
    }
  }
}

sub tosyslog {        # return void
  use Sys::Syslog;
  my $lvl = shift; #Sysloglevel
  my $msg = shift; #debug message
  openlog('YAFP', 'cons,pid', $FACILITY);
  syslog($lvl,$msg);
  closelog();
}

sub setup_defaults {   # return void
  #set some defaults
  $HOST         = '127.0.0.1';
  $PORT         = 1712;
  $SLAVE_PORT   = 15;
  $PIDFILE      = '/tmp/yafp.pid';
  $TIMEOUT      = 30;
  $DEBUGLEVEL   = 3;  
  $SYSLOG       = 1;
  $FACILITY     = 'daemon';
}

__END__

=pod

=head1 INTRODUCTION

YAFP is a proxy to filter out advertising banners and malicious script code
(read: popups) from 
Websites. It is written in Perl and based on an example by Randal L. Schwarz 
L<http://www.webtechniques.com/archives/1999/02/perl/>. For performance 
increasements it uses preforking to spawn several processes of it self (like 
Apache does). Unlike other available proxies (e.g. Junkbuster) it does blocking 
not only by URL patterns but by Perl regular expressions on the 
sites content, too.

The newest version can always be found at L<http://www.splitbrain.org>

=head1 REQUIREMENTS

YAFP utilizes some perlmodules which are HTTP::Daemon, HTTP::Request,
HTTP::Response and LWP::UserAgent which are all found in the libwww-perl bundle.
For parsing the HTML it uses the HTML::TreeBuilder module which is part of
the HTML::Tree distribution. To work with gzip encoded HTML content send by
modern Webserverinstallations Compress::Zlib is needed.

You may install them via the CPAN by:

 perl -MCPAN -e shell
 install Bundle::LWP
 install Compress::Zlib

Debian users may use
C<apt-get install libwww-perl libhtml-tree-perl libcompress-zlib-perl> as well.

Please note, bannerfiltering is a time and memory consuming task so you should 
have a machine that is able to handle that. I've run YAFP on a Pentium 266Mhz 
with 64MB RAM for my private homenetwork with acceptable speed, but you should 
consider using a better machine than that.


=head1 INSTALLATION

There is no sophisticated install script yet - so here's a short manual 
description:

 cp yafp.pl /usr/local/bin/
 mkdir /etc/yafp
 cp *.conf /etc/yafp/
 vi /etc/yafp/yafp.conf

You may use the provided F<yafp.init> file for running yafp on startup, too. 

=head1 CONFIGURATION

YAFP looks in different places for its main configuration file in the following 
order:

 /etc/yafp.conf
 /etc/yafp/yafp.conf
 ~/.yafp.conf
 ./yafp.conf

A later one overwrites a previous one. So you can have system wide configs which 
can be overwritten by each user.

To force all childs to reread the configfiles send a SIGHUP to one of them
(kill -1 <PID>) - B<Note:> This will I<not> reload the main
configuration file!


=head2 How the filters work

First a general overview on how YAFP will request and filter websites for you.
It is important to understand in which order things are done to configure the
proxy in the right way.

 Browser |                 YAFP               |  external data
 side    |                                    |
        -->  1. change Useragentstring and   <-- yafp.conf
         |      remove FROM header            |
         |                                    |
         |   2. if URL is a trusted partner  <-- trusted.conf
         |      -> 5                          |
         |                                    |
         |   3  is URL a cookie URL          <-- cookie.conf
         |      -> 5                          |
         |                                    |
         |   4  remove cookies from request   |
         |                                    |
         |   5. fetch Request                <-> WWW
         |                                    |
         |   6. if servercode != 200          |
        <--     return serveranswer           |
         |                                    |
         |   7. if mimetype != text/html      |
        <--     return content                |
         |                                    |
         |   8. if URL is a trusted partner  <-- yafp.conf
        <--     return serveranswer           |
         |                                    |
         |   9. if $FILTER_SCRIPT            <-- yafp.conf
         |      -> remove all JavaScript      |
         |      -> 13                         |
         |                                    |
         |  10. if $FILTER_ONLOAD            <-- yafp.conf
         |      -> remove onLoad and onUnload |
         |         JavaScript events          |
         |                                    |
         |  11. if $FILTER_STATUS            <-- yafp.conf
         |      -> remove stauschanging JS    |
         |                                    |
         |  12. if $FILTER_POPUP             <-- yafp.conf
         |      -> remove open() from all JS  |
         |                                    |
         |  13. if $FILTER_SIZE              <-- yafp.conf
         |      -> remove all IMG tags with   |
         |         banner sizes              <-- yafp.conf
         |                                    |
         |  14. kill all tags that use a     <-- blockurls.conf
         |      blockurl as SRC               |
         |                                    |
         |  15. apply replacement patterns   <-- patterns.conf
         |                                    |
        <-- return content                    |

=head2 Simplified Regular Expressions

The following files may contain patterns based on slightly simplyfied Perl 
regular expressions. The difference to real Pearl regexps is the handling of 
some special chars and asterisks (*). The following characters are automatically 
escaped: . @ % $ { } / - So note that a dot is not a special char anymore. 
Asterisks are expanded to a minimalistik "catch all" (.*?).

=head2 Trusted URLs

Trusted URLs are websites that you want to never have be filtered. This may be
for example your own site or a site where all YAFP kills to much. These URLs
are defined in the F<trusted.conf> config file.

=head2 Cookie URLs

These are similar to the trusted URLs but here the sites still will be filtered
but cookies are allowed. Cookie URLs are stored in the F<cookies.conf> file

=head2 Blocked URLs

In the file F<blockurls.conf> you may define patterns to determine advertising.
YAFP will check the following HTML tags for source URLs that match these patterns:
img, iframe, layer, embed, applet and object. If a pattern matches the tag is
removed from the sites content.

YAFP comes with a small tool called F<testurl.pl> to test your F<blockurl.conf>.
Give an URL as parameter to see if it is blocked by your current blockurl
configuration. 

=head2 Patterns

In addition to the above method of removing advertisements you can substitute parts
of the site by applying your own patterns. These are stored in the F<patterns.conf>
file. Give a pattern-replacement-pair on each line separated by a single tabulator
character.


=head1 FEEDBACK

Yes please! I need as much help as I possible - send in everything: patches,
comments, new or fixed patterns, money ;-)

Send your mail to a.gohr@web.de


=head1 LICENSE

YAFP is based upon a proxy written by Randal L. Schwartz which is licensed under 
"the same terms as Perl itself" - As Perl itself is licensed under GPL I<or> 
Artistic License I've choosen the GPL for this software. So here comes the 
disclaimer:

YAFP - Yet another filter proxy
Copyright (C) 2001,2002 Andreas Gohr <a.gohr@web.de>

This program is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; either version 2 of the License, or (at your option) any later 
version.

See COPYING for details

=cut
